import json
from pathlib import Path
from pdb import set_trace
from typing import Generator

import geopandas as gpd
import numpy as np
from scipy.spatial import Voronoi
from shapely.geometry import LineString, MultiPolygon, Polygon
from shapely.ops import polygonize
from voronoi_mapper.geojson import (
    load_mask_geojson,
    load_points_and_features_from_geojson,
)
from voronoi_mapper.geometry import (
    clip_polygons_to_mask,
    get_bounding_segments,
    get_intersection_with_bounding_box,
    match_point_features_to_polygons,
)
from voronoi_mapper.models import BoundingBox, Edges
from voronoi_mapper.plot import plot_voronoi


def _remove_coordinate_duplicates_from_list(coordinate_list: list[list[float]]):
    """Remove duplicate coordinates from list."""
    unique_tuples = set(tuple(item) for item in coordinate_list)
    return [list(item) for item in unique_tuples]


def _remove_segment_duplicates_from_list(segments: list[list[list[float]]]):
    """Remove duplicate segments from list.

    Segment AB is treated as different to segment BA.
    """
    unique_segments = set(tuple(map(tuple, segment)) for segment in segments)
    return [list(map(list, segment)) for segment in unique_segments]


def get_line_segments_from_voronoi(
    voronoi: Voronoi, bounding_box: BoundingBox
) -> tuple[list[list[float]], dict[str, list[list[float]]]]:
    """Get the list of line segments within a bounding box from a scipy Voronoi object."""
    center = voronoi.points.mean(axis=0)
    segments: list[tuple[float, float]] = []
    bounding_box_intersections: dict[str, list[list[float]]] = {
        Edges.top.value: [],
        Edges.right.value: [],
        Edges.bottom.value: [],
        Edges.left.value: [],
    }
    # Essentially a copy of the code from voronoi_plot_2d but with extension of
    # infinity edges to a bounding box.
    for pointidx, simplex in zip(voronoi.ridge_points, voronoi.ridge_vertices):
        simplex = np.asarray(simplex)
        if np.all(simplex >= 0):
            segments.append(voronoi.vertices[simplex].tolist())
            continue

        i = simplex[simplex >= 0][0]

        t = voronoi.points[pointidx[1]] - voronoi.points[pointidx[0]]
        t /= np.linalg.norm(t)
        n = np.array([-t[1], t[0]])

        midpoint = voronoi.points[pointidx].mean(axis=0)
        direction = np.sign(np.dot(midpoint - center, n)) * n

        # intersection with bounding box
        intersection = get_intersection_with_bounding_box(
            coordinates=(voronoi.vertices[i], voronoi.vertices[i] + direction),
            bounding_box=bounding_box,
        )

        segments.append([voronoi.vertices[i], intersection.coordinates])

        bounding_box_intersections[intersection.edge.value].append(
            intersection.coordinates
        )

    segments = _remove_segment_duplicates_from_list(segments=segments)

    for k, v in bounding_box_intersections.items():
        bounding_box_intersections[k] = _remove_coordinate_duplicates_from_list(
            coordinate_list=v
        )

    return segments, bounding_box_intersections


def get_polygons_from_voronoi(
    voronoi: Voronoi, bounding_box: BoundingBox
) -> Generator[Polygon, None, None]:
    segments, bounding_box_intersections = get_line_segments_from_voronoi(
        voronoi=voronoi, bounding_box=bounding_box
    )

    bounding_segments = get_bounding_segments(
        bounding_box=bounding_box, bounding_box_intersections=bounding_box_intersections
    )

    segments.extend(bounding_segments)

    segments = [LineString(coordinates=segment) for segment in segments]

    return polygonize(segments)


def create_geodataframe_from_polygons_and_features(
    matched_polygons_and_features: list[tuple[Polygon, dict]]
) -> gpd.GeoDataFrame:
    gdf = gpd.GeoDataFrame(
        [
            {"geometry": matched_polygon, **matched_feature["properties"]}
            for matched_polygon, matched_feature in matched_polygons_and_features
        ]
    )
    gdf = gdf.set_geometry("geometry")
    return gdf


def get_bounding_box(voronoi: Voronoi, mask: Polygon | MultiPolygon) -> BoundingBox:
    finite_vertices = voronoi.vertices[np.all(np.isfinite(voronoi.vertices), axis=1)]

    min_x, min_y = np.min(finite_vertices, axis=0)
    max_x, max_y = np.max(finite_vertices, axis=0)

    mask_min_x, mask_min_y, mask_max_x, mask_max_y = mask.bounds

    min_x = min(min_x, mask_min_x)
    min_y = min(min_y, mask_min_y)

    max_x = max(max_x, mask_max_x)
    max_y = max(max_y, mask_max_y)

    return BoundingBox(
        xmin=min_x - 1,
        xmax=max_x + 1,
        ymin=-min_y - 1,
        ymax=max_y + 1,
    )


def voronoi_map(
    features_file_path: Path | str,
    boundary_file_path: Path | str,
    bounding_box: BoundingBox | None = None,
    save_path: Path | str | None = None,
):
    points, features = load_points_and_features_from_geojson(
        geojson_path=features_file_path
    )
    mask = load_mask_geojson(geojson_path=boundary_file_path)

    voronoi = Voronoi(points)

    bounding_box = (
        get_bounding_box(voronoi=voronoi, mask=mask)
        if bounding_box is None
        else bounding_box
    )

    polygons = get_polygons_from_voronoi(voronoi=voronoi, bounding_box=bounding_box)

    matched_points = match_point_features_to_polygons(
        polygons=polygons, features=features
    )

    gdf = create_geodataframe_from_polygons_and_features(
        matched_polygons_and_features=matched_points
    )

    gdf = clip_polygons_to_mask(gdf=gdf, mask=mask)

    if save_path is not None:
        gdf.to_file(save_path, driver="GeoJSON")



from dataclasses import dataclass
from enum import Enum


class Edges(Enum):
    top = "top"
    right = "right"
    bottom = "bottom"
    left = "left"


@dataclass
class Intersection:
    coordinates: tuple[float, float]
    edge: Edges


@dataclass
class BoundingBox:
    xmin: float
    xmax: float
    ymin: float
    ymax: float


class IntersectionException(BaseException): ...



import json
from pathlib import Path

from shapely import get_coordinates
from shapely.geometry import shape
from shapely.ops import unary_union


def _load_geojson_file(geojson_path: str | Path):
    # TODO: extend to yaml
    with open(geojson_path, "r", encoding="utf-8") as f:
        points_geojson_obj = json.load(f)
    return points_geojson_obj


def load_points_and_features_from_geojson(
    geojson_path: str | Path,
) -> tuple[list[list[float]], list[dict[str, str]]]:
    geojson_obj = _load_geojson_file(geojson_path=geojson_path)

    features = geojson_obj["features"]
    geom = [shape(i["geometry"]) for i in features]
    points = [get_coordinates(x).tolist() for x in geom]
    points = [item[0] for item in points]

    return points, features


def load_mask_geojson(geojson_path: str | Path):
    boundary_geojson_obj = _load_geojson_file(geojson_path=geojson_path)

    geometries = [shape(x["geometry"]) for x in boundary_geojson_obj["features"]]

    unified_shape = unary_union(geoms=geometries)
    return unified_shape



from typing import Generator, Literal

import geopandas as gpd
from shapely.geometry import MultiPolygon, Polygon, shape
from voronoi_mapper.models import (
    BoundingBox,
    Edges,
    Intersection,
    IntersectionException,
)


def calculate_gradient(point_a: list[float], point_b: list[float]) -> float | None:
    """Calculate the gradient of the line segment between point A and point B."""
    try:
        return (point_b[1] - point_a[1]) / (point_b[0] - point_a[0])
    except ZeroDivisionError:
        # vertical line
        return None


def check_point_is_within(
    point_a: list[float], point_b: list[float], point_c: list[float]
):
    """Checks if point_b is between point_a and point_c.

    Assumes they are colinear.
    """
    return (
        min(point_a[0], point_c[0]) <= point_b[0] <= max(point_a[0], point_c[0])
    ) and (min(point_a[1], point_c[1]) <= point_b[1] <= max(point_a[1], point_c[1]))


def handle_straight_lines(
    point_a: list[float],
    point_b: list[float],
    gradient: Literal[0] | None,
    bounding_box: BoundingBox,
):
    if gradient == 0:
        # horizontal
        left_to_right = (point_b[0] - point_a[0]) > 0
        return Intersection(
            coordinates=(
                bounding_box.xmax if left_to_right else bounding_box.xmin,
                point_a[1],
            ),
            edge=Edges.right if left_to_right else Edges.left,
        )
    # vertical
    bottom_to_top = (point_b[1] - point_a[1]) > 0
    return Intersection(
        coordinates=(
            point_b[0],
            bounding_box.ymax if bottom_to_top else bounding_box.ymin,
        ),
        edge=Edges.top if bottom_to_top else Edges.bottom,
    )


def calculate_slanted_intersections(
    point_a: list[float], gradient: float, bounding_box: BoundingBox
) -> list[Intersection]:
    intersections: list[Intersection] = []
    c_constant = point_a[1] - gradient * point_a[0]

    # check intersection with each side of the bounding box
    # left side (x = xmin)
    y = gradient * bounding_box.xmin + c_constant
    if bounding_box.ymin <= y <= bounding_box.ymax:
        intersections.append(
            Intersection(coordinates=(bounding_box.xmin, y), edge=Edges.left)
        )

    # right side (x = xmax)
    y = gradient * bounding_box.xmax + c_constant
    if bounding_box.ymin <= y <= bounding_box.ymax:
        intersections.append(
            Intersection(coordinates=(bounding_box.xmax, y), edge=Edges.right)
        )

    # bottom side (y = ymin)
    x = (bounding_box.ymin - c_constant) / gradient
    if bounding_box.xmin <= x <= bounding_box.xmax:
        intersections.append(
            Intersection(coordinates=(x, bounding_box.ymin), edge=Edges.bottom)
        )

    # top side (y = ymax)
    x = (bounding_box.ymax - c_constant) / gradient
    if bounding_box.xmin <= x <= bounding_box.xmax:
        intersections.append(
            Intersection(coordinates=(x, bounding_box.ymax), edge=Edges.top)
        )
    return intersections


def find_main_intersection(
    point_a, point_b, intersections: list[Intersection]
) -> Intersection | None:
    main_intersection: Intersection | None = None
    for intersection in intersections:
        if check_point_is_within(
            point_a=point_a, point_b=point_b, point_c=intersection.coordinates
        ):
            main_intersection = intersection
    return main_intersection


def get_intersection_with_bounding_box(
    coordinates: tuple[list[float], list[float]], bounding_box: BoundingBox
) -> Intersection:
    """
    Calculate the intersection of a line segment defined by two points with a bounding box.

    Parameters:
    - coordinates (tuple[list[float], list[float]]): A tuple containing two lists representing point A and point B.
    - bounding_box (BoundingBox): An instance of BoundingBox defining the area of interest.

    Returns:
    - Intersection: An instance of Intersection containing the coordinates of the intersection and the edge it intersects with.

    Raises:
    - IntersectionException: If no intersection is found within the bounding box.
    """
    point_a, point_b = coordinates
    gradient = calculate_gradient(point_a, point_b)

    # handle horizontal and vertical lines
    if gradient in [0, None]:
        return handle_straight_lines(point_a, point_b, gradient, bounding_box)

    intersections = calculate_slanted_intersections(point_a, gradient, bounding_box)
    main_intersection = find_main_intersection(point_a, point_b, intersections)

    if main_intersection is None:
        raise IntersectionException(
            f"No intersections found for coordinates {coordinates} in this bounding box {bounding_box}."
        )
    return main_intersection


def get_bounding_segments(
    bounding_box: BoundingBox, bounding_box_intersections: dict[str, list]
):
    bounding_segments = []
    for k, list_of_intersections in bounding_box_intersections.items():
        if k in [Edges.top.value, Edges.bottom.value]:
            y_val = bounding_box.ymax if k == Edges.top.value else bounding_box.ymin
            start_point = (
                bounding_box.xmin,
                y_val,
            )
            end_point = (
                bounding_box.xmax,
                y_val,
            )

            list_of_intersections = sorted(list_of_intersections)
            list_of_intersections.insert(0, start_point)
            list_of_intersections.append(end_point)
        else:
            x_val = bounding_box.xmax if k == Edges.right.value else bounding_box.xmin
            start_point = (
                x_val,
                bounding_box.ymin,
            )
            end_point = (
                x_val,
                bounding_box.ymax,
            )
            list_of_intersections.sort(key=lambda x: x[1])
            list_of_intersections.insert(0, start_point)
            list_of_intersections.append(end_point)

        bounding_segments.extend(
            [
                (list_of_intersections[i], list_of_intersections[i + 1])
                for i in range(len(list_of_intersections) - 1)
            ]
        )
    return bounding_segments


def match_point_features_to_polygons(
    polygons: list[Polygon] | Generator[Polygon, None, None], features: list
) -> list[tuple[Polygon, dict]]:
    polygon_point_pairs = []
    for feature in features:
        for polygon in polygons:
            if not polygon.contains(shape(feature["geometry"])):
                continue
            polygon_point_pairs.append((polygon, feature))
            break
    return polygon_point_pairs


def clip_polygons_to_mask(
    gdf: gpd.GeoDataFrame, mask: MultiPolygon
) -> gpd.GeoDataFrame:
    return gdf.clip(mask=mask)



from pathlib import Path
from pdb import set_trace

import geopandas as gpd
import matplotlib.pyplot as plt
from matplotlib import patches
from matplotlib.axes import Axes
from scipy.spatial import Voronoi, voronoi_plot_2d
from shapely.geometry import MultiPolygon, Polygon
from voronoi_mapper.models import BoundingBox


def plot_voronoi(
    voronoi: Voronoi,
    save_path: Path | str,
    bounding_box: BoundingBox | None = None,
    boundary: Polygon | MultiPolygon | None = None,
    show: bool = False,
):
    _, ax = plt.subplots(figsize=(8, 6))
    ax: Axes = ax

    # gridlines
    ax.grid(which="major", color="#BBBBBB", linewidth=0.6)
    ax.grid(which="minor", color="#CCCCCC", linestyle=":", linewidth=0.35)
    ax.minorticks_on()

    voronoi_plot_2d(
        voronoi,
        ax=ax,
        show_vertices=True,
        line_colors="blue",
        line_width=2,
        point_size=5,
    )

    # bounding box
    if bounding_box is not None:
        width = bounding_box.xmax - bounding_box.xmin
        height = bounding_box.ymax - bounding_box.ymin
        rectangle = patches.Rectangle(
            (bounding_box.xmin, bounding_box.ymin),
            width,
            height,
            linewidth=1,
            edgecolor="#39fc03",
            facecolor="none",
        )
        ax.add_patch(rectangle)

        ax.set_xlim((bounding_box.xmin - 1, bounding_box.xmax + 1))
        ax.set_ylim((bounding_box.ymin - 1, bounding_box.ymax + 1))

    # boundary
    if boundary is not None:
        if boundary.geom_type == "Polygon":
            plot_polygon(ax, boundary)

        elif boundary.geom_type == "MultiPolygon":
            for polygon in boundary.geoms:
                plot_polygon(ax, polygon)

    # titles
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    ax.set_title("Voronoi Diagram")

    plt.savefig(save_path)
    if show:
        plt.show()  # pragma: no cover


def plot_polygon(ax: Axes, polygon: Polygon):
    """
    Helper function to plot a single polygon.

    Parameters:
        ax (matplotlib axes): The axes on which to plot.
        polygon (Shapely Polygon): The polygon to plot.
    """
    x, y = polygon.exterior.xy
    ax.fill(x, y, alpha=0.5, color="orange")  # Fill the polygon with transparency
